/* $Id: acre_effects.c,v 1.1 2004/04/30 05:15:50 Joel M. Malard Exp d3j540 $ */

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "acre.h"
#include <assert.h>
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_effects"
/* Compute the effects in the linear model and
 * return them in the sequential vector x.
 */
int acre_effects( acre_cntxt *cntxt, Vec x, PetscScalar *rnorm)
{
	int n = cntxt->n;
        int rn = cntxt->rn;

	int rise, rank, cank;
	int *dis, *off, *rows, nrows, start, end;
	int i, ierr;

	PetscScalar *qVals;
	PetscScalar *y1Vals, *y2Vals, *y3Vals;
	PetscScalar *z1Vals;

	Vec y1;
	Vec y2;
	Vec res;

	KSP ksp;

	/* constants */
	int ione = 1;
	PetscScalar aone = 1.0;
	PetscScalar azero = 0.0;
	PetscScalar amone = -1.0;

	HELLO("acre_effects enter");

	/* rank in process row */
	ierr = MPI_Comm_rank(cntxt->rcomm,&rank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_size(cntxt->rcomm,&rise);
		assert(ierr==MPI_SUCCESS);

	/* rank in process column */
	ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&cank);
		assert(ierr==MPI_SUCCESS);

	/* Allocate memory */
	ierr = VecDuplicate(cntxt->b,&y1);		CHKERRQ(ierr);

	/* Find the range of local rows */
	ierr = VecGetOwnershipRange (cntxt->b,
		&start, &end);				CHKERRQ(ierr);
	nrows = end-start;

	/* y1 = w-v */
	ierr = VecCopy(cntxt->w, y1);			CHKERRQ(ierr);
	ierr = VecAXPY(&amone, cntxt->v, y1);		CHKERRQ(ierr);

	/* y2 = Q'y1 */
	/* Compute local partial sums for y2 */
	ierr = MatGetArray( cntxt->Q, &qVals );		CHKERRQ(ierr);
	ierr = VecGetArray( y1, &y1Vals );		CHKERRQ(ierr);
	ierr = PetscMalloc( rn*sizeof(PetscScalar),
		&y2Vals);				CHKERRQ(ierr);
	ierr = PetscMalloc( rn*sizeof(PetscScalar),
		&y3Vals);				CHKERRQ(ierr);

	if ( nrows==0 )
	{
		for ( i=0; i<n; i++ ) y2Vals[i] = 0.0;
	}
	else
	{
		LAgemv_("Tr", &nrows, &rn, &aone,
			qVals, &nrows, y1Vals, &ione,
			&azero, y2Vals, &ione);		CHKERRQ(ierr);
	}
	/* Compute global y2 components within process
	 * column.
	 */
	ierr = MPI_Allreduce(y2Vals, y3Vals, rn,
		MPI_DOUBLE, MPI_SUM, PETSC_COMM_WORLD);
		assert(ierr==MPI_SUCCESS);
	PetscFree(y2Vals);
	ierr = VecRestoreArray( y1, &y1Vals );		CHKERRQ(ierr);
	ierr = MatRestoreArray( cntxt->Q, &qVals );	CHKERRQ(ierr);

	/* Gather y2 components across process columns */
	ierr = PetscMalloc( rise*sizeof(int), &off);	CHKERRQ(ierr);
	ierr = PetscMalloc( rise*sizeof(int), &dis);	CHKERRQ(ierr);
	ierr = PetscMalloc( n*sizeof(PetscScalar),
		&z1Vals);				CHKERRQ(ierr);
	for ( i=0; i<rise; i++ )
		off[i] = cntxt->rstart[i+1] -
			cntxt->rstart[i];
	dis[0] = 0;
	for ( i=1; i<rise; i++ )
		dis[i] = off[i] + dis[i-1];
	ierr = MPI_Allgatherv(y3Vals, rn, MPI_DOUBLE,
		z1Vals, off, dis, MPI_DOUBLE,
		cntxt->rcomm);
		assert(ierr==MPI_SUCCESS);

	PetscFree(off);
	PetscFree(dis);

	ierr = VecCreateSeq(PETSC_COMM_SELF, n, &y2);	CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(int), &rows);	CHKERRQ(ierr);
	for ( i=0; i<n; i++ ) rows[i] = i;		CHKERRQ(ierr);
	ierr = VecSetValues( y2, n, rows, z1Vals,
		INSERT_VALUES);				CHKERRQ(ierr);
	ierr = VecAssemblyBegin(y2);			CHKERRQ(ierr);
	ierr = VecAssemblyEnd(y2);			CHKERRQ(ierr);
	PetscFree(rows);
	PetscFree(z1Vals);
	PetscFree(y3Vals);

	/* Rx = y2 */

	ierr = KSPCreate(PETSC_COMM_SELF, &ksp);	CHKERRQ(ierr);
	ierr = KSPSetOperators (ksp, cntxt->R,
		cntxt->R, DIFFERENT_NONZERO_PATTERN);	CHKERRQ(ierr);
	ierr = KSPSetRhs(ksp, y2);			CHKERRQ(ierr);
	ierr = KSPSetSolution(ksp, x);			CHKERRQ(ierr);
	ierr = KSPSetFromOptions(ksp);			CHKERRQ(ierr);

	ierr = KSPSolve(ksp);				CHKERRQ(ierr);
	ierr = KSPDestroy(ksp);				CHKERRQ(ierr);

	/* Compute residual norm */
	ierr = VecDuplicate(cntxt->b,&res);		CHKERRQ(ierr);
	ierr = MatMult( cntxt->L, cntxt->v, res);	CHKERRQ(ierr);
	ierr = VecNorm(res,NORM_2, rnorm);		CHKERRQ(ierr);

	/* Clean up */
	ierr = VecDestroy(y1);				CHKERRQ(ierr);
	ierr = VecDestroy(y2);				CHKERRQ(ierr);
	ierr = VecDestroy(res);				CHKERRQ(ierr);

	HELLO("acre_effects exit");

	return 0;
}
