/*$Id: acre_sles_mat.c,v 1.1 2004/04/30 Joel M. Malard Exp d3j540 $ */

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "petsclog.h"
#include "acre.h"
#include "petsclog.h"
#include <assert.h>
#include <mpi.h>
#include <math.h>
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_sles_mat"
int acre_sles_mat( acre_cntxt *cntxt, Mat A, Mat B, Vec v, int bTranspose)
{
	/* Purpose:

	Solve multiple systems of linear equations, this procedure is called
	with serial and parallel matrices. The vector v serves as a template
	to allocate workspace vectors. It is not modified. The alternative
	seems to be to check the type of matrix A and B against all possible
	matrix types.

	*/
	Vec aVec;
	Vec bVec;
	int i, ierr = 0;
	int nrows, *rows;
	int start, end;
	int m, k, h, rn;
	PetscScalar *vals;

	HELLO("acre_sles_mat enter");	

	ierr = MatGetSize( A, &m, &rn );			CHKERRQ(ierr);
	ierr = MatGetSize( B, &h, &k );				CHKERRQ(ierr);
	if(h!=m || k!=rn)
	{
		PetscPrintf(PETSC_COMM_WORLD,
		"Bm=%d\tBn=%d\tAm=%d\tAn=%d\n",h,k,m,rn);
		CHKERRQ(-1);
	}

	/* test */
	if ( bTranspose==ACRE_TRANSPOSE )
	{
		ierr = KSPSetInitialGuessNonzero( cntxt->stes,
			PETSC_TRUE );				CHKERRQ(ierr);
	}
	else
	{
		ierr = KSPSetInitialGuessNonzero( cntxt->sles,
			PETSC_TRUE );				CHKERRQ(ierr);
	}

	ierr = VecDuplicate( v, &aVec);				CHKERRQ(ierr);
	ierr = VecDuplicate( v, &bVec);				CHKERRQ(ierr);
	ierr = VecGetOwnershipRange( aVec, &start, &end );	CHKERRQ(ierr);
	nrows = end-start;
	rows = NULL;
	if ( nrows!=0 )
	{
		ierr = PetscMalloc(nrows*sizeof(int),&rows);	CHKERRQ(ierr);
		for ( i=0; i<nrows; i++) rows[i] = i+start;
	}
	for ( i=0; i<rn; i++ )
	{
		ierr = acre_get_dense_matrix_column_vector (
			A, aVec, i );				CHKERRQ(ierr);
		ierr = acre_get_dense_matrix_column_vector (
			B, bVec, i );				CHKERRQ(ierr);
		if ( bTranspose==ACRE_TRANSPOSE )
		{
			ierr = KSPSetRhs(cntxt->stes, aVec);	CHKERRQ(ierr);
			ierr = KSPSetSolution(cntxt->stes,
				bVec);				CHKERRQ(ierr);
			ierr = KSPSolve( cntxt->stes ) ; 	CHKERRQ(ierr);
		}
		else
		{
			ierr = KSPSetRhs(cntxt->sles, aVec);	CHKERRQ(ierr);
			ierr = KSPSetSolution(cntxt->sles,
				bVec);				CHKERRQ(ierr);
			ierr = KSPSolve( cntxt->sles );		CHKERRQ(ierr);
		}
		if ( nrows!=0 )
		{
			ierr = VecGetArray ( bVec, &vals );	CHKERRQ(ierr);
			ierr = MatSetValues( B, nrows,
				rows, 1, &i, vals,
				INSERT_VALUES);			CHKERRQ(ierr);
			ierr = VecRestoreArray ( bVec, &vals );	CHKERRQ(ierr);
		}
		else
		{
			int rank;
  			ierr = MPI_Comm_rank(
				PETSC_COMM_WORLD,&rank);
				assert(ierr==MPI_SUCCESS);
			PetscPrintf(PETSC_COMM_SELF,
				"Process %d doesn't own B\n",
				rank);
		}
	}
	ierr = MatAssemblyBegin( B, MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
	ierr = MatAssemblyEnd( B, MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);
	ierr = VecDestroy ( aVec );				CHKERRQ(ierr);
	ierr = VecDestroy ( bVec );				CHKERRQ(ierr);
	if ( nrows!=0 ) PetscFree(rows);

	HELLO("acre_sles_mat exit");	

	return 0;

}

